package no.tornado.query;

/**
 * <p>A JoinTable is used in a Mapper to configure how a table should be
 * joined into a query. The minimal requirement for a JoinTable instance
 * is that you tell it the table name, an alias to use, and how to perform
 * the join. There is no magic here, you must spesifically write the correct
 * SQL to use for the join - or let the Mapper.generate() method create
 * a Mapper with JoinTables for you.</p>
 *
 * <p>To join a customer with an address:</p>
 *
 * <pre>
 * JoinTable address = new JoinTable("address", "a").on("a.id = customer.address")
 * </pre>
 *
 * <p>When used in a CustomerMapper, the JoinTable instance from above would generate the following SQL:</p>
 *
 * <pre>
 * SELECT &lt;fields&gt; FROM customer, address a WHERE a.id = customer.address
 * </pre>
 *
 * <p>If you want to create a left join instead, you can instantate the JoinTable like this:</p>
 *
 * <pre>
 * JoinTable address = new JoinTable("address", "a", LEFT_JOIN).on("a.id = customer.address")
 * </pre>
 *
 * <p>In the customer Mapper you would use the address JoinTable once for each column to join in,
 * using the <code>.join()</code> method:</p>
 *
 * <pre>
 * Mapper<Customer> customerMapper = new Mapper(Customer.class)
 *  .tablename("customer")
 *  .property("address.id", "address") // Set the address column of customer to customer.address.id
 *  .join("address.street", address, "street") // Join the street column of address to customer.address.street
 *  .join("address.zip", address, "zip") // Join the zip column of address to customer.address.zip
 * </pre>
 *
 * A JoinTable instance can be shared across many mappers.
 *
 * @see JOIN_TYPE
 * @see Mapper#generate(Class domainClass, String targetPackage, String tablename)
 */
public class JoinTable {
    /**
     * The JOIN_TYPE enum will tell a JoinTable how to perform the actual join.
     * <ul>
     *     <li><code>INNER_JOIN</code> is default, and results in a JOIN</li>
     *     <li><code>OUTER_JOIN</code> results in LEFT OUTER JOIN</li>
     * </ul>
     *
     * The <code>.on()</code> joining expression is used in both cases.
     */
    public enum JOIN_TYPE { INNER_JOIN, OUTER_JOIN };
    
    private String name;
    private String alias;
    private JOIN_TYPE joinType;
    private String on;
    public transient String pkColumn;

    public JoinTable(String name) {
        this(name, name, JOIN_TYPE.INNER_JOIN);
    }

    public JoinTable(String name, String alias) {
        this(name, alias, JOIN_TYPE.INNER_JOIN);
    }
    
    public JoinTable(String name, String alias, JOIN_TYPE joinType) {
        this.name = name;
        this.alias = alias;
        this.joinType = joinType;
    }

    public String toString() {
        return joinType + " " + name + " AS " + alias;
    }

    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        JoinTable joinTable = (JoinTable) o;

        if (!alias.equals(joinTable.alias)) return false;
        if (joinType != joinTable.joinType) return false;
        if (!name.equals(joinTable.name)) return false;

        return true;
    }

    public int hashCode() {
        int result = name.hashCode();
        result = 31 * result + alias.hashCode();
        result = 31 * result + joinType.hashCode();
        return result;
    }

    public JOIN_TYPE getJoinType() {
        return joinType;
    }

    public JoinTable joinType(JOIN_TYPE joinType) {
        this.joinType = joinType;
        return this;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getAlias() {
        return alias;
    }

    public void setAlias(String alias) {
        this.alias = alias;
    }

    public String getOn() {
        return on;
    }

    public void setOn(String on) {
        this.on = on;
    }

    public JoinTable on(String on) {
        this.on = on;
        return this;
    }

}
