package no.tornado.query;

import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Types;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class Mapper<T> implements Serializable {
    public enum MapType {PRIMARY_KEY, TABLE_COLUMN, JOIN_COLUMN}

    private Class resultClass;
    private Map<String, Mapping> mappings = new LinkedHashMap<>();
    public static final Boolean DEFAULT_STRICT = true;
    public static final Boolean DEFAULT_AUTOMAP = false;
    private Boolean strict = DEFAULT_STRICT;
    private Boolean automap = DEFAULT_AUTOMAP;
    private String tablename;

    private Mapper() { }

    public String describe() {
        StringBuilder s = new StringBuilder();
        s.append("Mapper " + resultClass.getName() + " for table " + tablename + ":\n\n");
        for (Mapping mapping : mappings.values())
            s.append("\t" + mapping.toString() + "\n");

        List<JoinTable> joinTables = uniqueJoinTables();
        if (!joinTables.isEmpty()) {
            s.append("\n");
            for (JoinTable joinTable : joinTables)
                s.append("\t" + joinTable + " ON " + joinTable.getOn() + "\n");
        }

        s.append("\n").append(select());
        s.append("\n").append(insert());
        s.append("\n").append(update());
        return s.toString();
    }

    public Map<String, Mapping> getMappings() {
        return mappings;
    }

    public static String generate(Class domainClass, String targetPackage, String tablename) throws SQLException, IllegalAccessException, InstantiationException, InvocationTargetException, NoSuchMethodException, IOException {
        return generate(domainClass, targetPackage, tablename, null);
    }

    public static String generate(Class domainClass, String targetPackage, String tablename, Path targetFolder) throws SQLException, IllegalAccessException, InstantiationException, InvocationTargetException, NoSuchMethodException, IOException {
        String className = tablename.substring(0, 1).toUpperCase() + tablename.substring(1).toLowerCase() + "Mapper";
        StringBuilder s = new StringBuilder("package " + targetPackage + ";\n\n");
        s.append("import " + domainClass.getName() + ";\n");
        s.append("import " + Mapper.class.getName() + ";\n");
        s.append("import " + JoinTable.class.getName() + ";\n");
        s.append("import static java.sql.Types.*;\n");
        s.append("\npublic class " + className + " {\n");
        
        ResultSet rs = Query.connection.get().createStatement().executeQuery("SELECT * FROM " + tablename);
        ResultSetMetaData m = rs.getMetaData();
        DatabaseMetaData dbm = Query.connection.get().getMetaData();

        ResultSet primaryKeys = dbm.getPrimaryKeys(null, null, tablename.toUpperCase());
        List<String> pks = new ArrayList<>();
        while (primaryKeys.next())
            pks.add(primaryKeys.getString("COLUMN_NAME").toLowerCase());

        ResultSet importedKeys = dbm.getImportedKeys(null, null, tablename.toUpperCase());
        Map<String, JoinTable> joinTables = new HashMap<>();
        while (importedKeys.next()) {
            String fkColumnName = importedKeys.getString("FKCOLUMN_NAME").toLowerCase();
            String pkTableName = importedKeys.getString("PKTABLE_NAME").toLowerCase();
            String pkColumnName = importedKeys.getString("PKCOLUMN_NAME").toLowerCase();
            JoinTable joinTable = new JoinTable(pkTableName, fkColumnName).on(tablename + "." + fkColumnName + " = " + fkColumnName + "." + pkColumnName);
            joinTable.pkColumn = pkColumnName;
            String camel = underscoreToCamelCase(fkColumnName);
            joinTables.put(camel, joinTable);
            s.append("\tprivate static final JoinTable " + camel + " = new JoinTable(")
                    .append("\"" + joinTable.getName() + "\", \"" + joinTable.getAlias() + "\").on(\"" + joinTable.getOn() + "\");\n");
        }
        if (!importedKeys.wasNull())
            s.append("\n");

        s.append("\tprivate " + className + "() {}\n\n");
        s.append("\t@SuppressWarnings(\"unchecked\")\n");
        s.append("\tpublic static final Mapper<" + domainClass.getSimpleName() +"> FULL = new Mapper(" + domainClass.getSimpleName() + ".class)\n");

        Object prototype = domainClass.newInstance();
        s.append("\t\t.tablename(\"").append(tablename).append("\")");
        for (int i = 0; i < m.getColumnCount(); i++) {
            String column = m.getColumnName(i + 1).toLowerCase();
            String property = setColumnProperty(prototype, column);
            JoinTable joinTable = joinTables.get(underscoreToCamelCase(column));
            Class nestedPropertyType = null;

            if (property != null && joinTable != null) {
                try {
                    Query.ensureObjectsCreated(prototype, property);
                    nestedPropertyType = Query.propertyUtils.getPropertyType(prototype, property);
                    String nestedProperty = property + "." + joinTable.pkColumn;
                    Query.ensureObjectsCreated(prototype, property);
                    Query.propertyUtils.getProperty(prototype, nestedProperty);
                    property = nestedProperty;
                } catch (Exception noSuchField) {
                    nestedPropertyType = null;
                }
            }

            String fieldType = pks.contains(column) ? "id" : "property";

            s.append("\n\t\t." + fieldType + "(\"").append(property != null ? property : "").append("\", \"").append(column).append("\"");
            ResultSet columnRs = dbm.getColumns(null, null, tablename.toUpperCase(), column.toUpperCase());
            boolean columnDataAvailable = columnRs.next();
            if (m.isAutoIncrement(i + 1) && columnDataAvailable) {
                String seqName = "sequence_name";
                if (columnRs.next()) {
                    String defaultValue = columnRs.getString("COLUMN_DEF");
                    Pattern seq = Pattern.compile(".*nextval\\(\'(.*?)\'.*");
                    Matcher matcher = seq.matcher(defaultValue);
                    if (matcher.matches())
                        seqName = matcher.group(1);
                }

                s.append(", \"" + seqName + "\"");
            }

            if (columnDataAvailable) {
                String typeName = columnRs.getString("TYPE_NAME");
                boolean typeSet = false;
                try {
                    if (Types.class.getDeclaredField(typeName) != null)
                        s.append(", ").append(typeName);
                    typeSet = true;
                } catch (Exception ex) {
                    
                }
                if (!typeSet)
                    s.append(", ").append(columnRs.getInt("DATA_TYPE"));
            }

            s.append(")");

            if (nestedPropertyType != null)
                appendJoinTableSkeleton(property.substring(0, property.lastIndexOf(".")), dbm, nestedPropertyType, s, joinTable);
        }
        s.append(";\n}\n");
        
        String generated = s.toString();
        if (targetFolder != null) {
            Path src = Paths.get(targetFolder.toFile().getAbsolutePath(), className + ".java");
            Files.write(src, generated.getBytes());
        }
        return generated;
    }

    private static String underscoreToCamelCase(String s) {
        Pattern us = Pattern.compile("_([a-z]{1})");
        Matcher m = us.matcher(s.toLowerCase());
        StringBuffer sb = new StringBuffer();
        while (m.find()) {
            String c = m.group(1).toUpperCase();
            m.appendReplacement(sb, c);
        }
        m.appendTail(sb);
        return sb.toString();
    }

    private static void appendJoinTableSkeleton(String parentProperty, DatabaseMetaData dbm, Class domainClass, StringBuilder s, JoinTable joinTable) throws SQLException, InvocationTargetException, NoSuchMethodException, IllegalAccessException, InstantiationException {
        ResultSet rs = Query.connection.get().createStatement().executeQuery("SELECT * FROM " + joinTable.getName());
        ResultSetMetaData m = rs.getMetaData();
        Object prototype = domainClass.newInstance();

        ResultSet importedKeys = dbm.getImportedKeys(null, null, joinTable.getName().toUpperCase());
        List<String> joinColumns = new ArrayList<>();
        while (importedKeys.next())
            joinColumns.add(importedKeys.getString("FKCOLUMN_NAME").toLowerCase());

        for (int i = 0; i < m.getColumnCount(); i++) {
            String column = m.getColumnName(i + 1).toLowerCase();

            if (column.equals(joinTable.pkColumn) || joinColumns.contains(column))
                continue;

            String property = setColumnProperty(prototype, column);
            String nestedProperty = parentProperty + "." + (property == null ? "" : property);

            s.append("\n\t\t.join(\"").append(nestedProperty).append("\", ").append(underscoreToCamelCase(joinTable.getAlias())).append(", \"").append(column).append("\"");

            ResultSet columnRs = dbm.getColumns(null, null, joinTable.getName().toUpperCase(), column.toUpperCase());
            if (columnRs.next()) {
                String typeName = columnRs.getString("TYPE_NAME");
                boolean typeSet = false;
                try {
                    if (Types.class.getDeclaredField(typeName) != null)
                        s.append(", ").append(typeName);
                    typeSet = true;
                } catch (Exception ex) {

                }
                if (!typeSet)
                    s.append(", ").append(columnRs.getInt("DATA_TYPE"));
            }

            s.append(")");
        }
    }

    private static String setColumnProperty(Object object, String property) {
        try {
            Query.propertyUtils.getProperty(object, property);
            return property;
        } catch (Exception ex) {
            if (property.contains("_")) {
                String fixedCase = underscoreToCamelCase(property);
                try {
                    Query.propertyUtils.getProperty(object, fixedCase);
                    return fixedCase;
                } catch (Exception e) {
                    return null;
                }
            }
        }
        return null;
    }

    public void merge(String columnPrefix, String propertyPrefix, Mapper<T> mapper) {
        for (String key : mapper.getMappings().keySet()) {
            Mapping mapping = mapper.getMappings().get(key);
            mapping.setProperty(propertyPrefix + "." + mapping.getProperty());
            mappings.put(columnPrefix + "_" + key, mapping);
        }
    }

    public String columns(Boolean includeJoins) {
        StringBuilder s = new StringBuilder();
        for (Mapping mapping : mappings.values()) {
            if (includeJoins || (mapping.getMapType().equals(MapType.TABLE_COLUMN) || mapping.getMapType().equals(MapType.PRIMARY_KEY))) {
                if (s.length() > 0)
                    s.append(", ");

                if (mapping.getJoinTable() != null)
                    s.append(mapping.getJoinTable().getAlias() + "." + mapping.getColumn() + " AS " + mapping.getJoinTable().getAlias() + "_" + mapping.getColumn());
                else if (includeJoins)
                    s.append(tablename + "." + mapping.getColumn());
                else
                    s.append(mapping.getColumn());
            }
        }
        return s.toString();
    }

    public String values() {
        StringBuilder s = new StringBuilder();
        for (Mapping mapping : mappings.values()) {
            if (mapping.getMapType().equals(MapType.TABLE_COLUMN) || mapping.getMapType().equals(MapType.PRIMARY_KEY)) {
                if (s.length() > 0)
                    s.append(", ");
                s.append(":").append(mapping.getProperty());
            }
        }
        return s.toString();
    }

    public String select() {
        StringBuilder s = new StringBuilder("SELECT " + columns(true) + " FROM " + tablename);
        List<JoinTable> joinTables = uniqueJoinTables();

        for (JoinTable joinTable : joinTables) {
            switch (joinTable.getJoinType()) {
                case INNER_JOIN: s.append(" JOIN "); break;
                case OUTER_JOIN: s.append(" LEFT OUTER JOIN "); break;
            }
            s.append(joinTable.getName()).append(" AS ").append(joinTable.getAlias()).append(" ON " + joinTable.getOn());
        }

        return s.toString();
    }

    public List<JoinTable> uniqueJoinTables() {
        List<JoinTable> joinTables = new ArrayList<>();
        for (Mapping mapping : mappings.values()) {
            if (mapping.getJoinTable() != null && !joinTables.contains(mapping.getJoinTable()))
                joinTables.add(mapping.getJoinTable());
        }
        return joinTables;
    }

    public String insert() {
        return "INSERT INTO " + tablename + " (" + columns(false) + ") VALUES (" + values() + ")";
    }

    public String updateStatement(String param) {
        StringBuilder s = new StringBuilder();
        for (Mapping mapping : mappings.values()) {
            if (mapping.getMapType().equals(MapType.TABLE_COLUMN)) {
                if (s.length() > 0)
                    s.append(", ");
                s.append(mapping.getColumn()).append(" = ").append(":");
                if (param != null)
                    s.append(param + ".");

                s.append(mapping.getProperty());
            }
        }
        return s.toString();
    }

    public String update() {
        return "UPDATE " + tablename + " SET " + updateStatement(null) + " WHERE " + identifyColumn();
    }

    private String identifyColumn() {
        StringBuilder s = new StringBuilder();
        for (Mapping mapping : mappings.values()) {
            if (mapping.getMapType().equals(MapType.PRIMARY_KEY)) {
                if (s.length() > 0)
                    s.append("AND ");
                s.append(mapping.getQualifiedColumn()).append("=:").append(mapping.getProperty());
            }
        }
        return s.toString();
    }

    public String update(String param) {
        return "UPDATE " + tablename + " SET " + updateStatement(param) + " WHERE " + identifyColumn();
    }

    public String delete() {
        StringBuilder s = new StringBuilder("DELETE FROM " + tablename + " WHERE ");
        boolean first = true;
        for (Mapping m : getIdMappings()) {
            if (first) {
                first = false;
            } else {
                s.append(" AND ");
            }
            s.append(m.getColumn() + " = :" + m.getProperty());
        }
        return s.toString();
    }

    
    public void setMappings(Map<String, Mapping> mappings) {
        this.mappings = mappings;
    }

    public boolean containsColumn(String columnName) {
        return mappings.containsKey(columnName);
    }

    public <T> Mapper(Class<T> resultClass) {
        this.resultClass = resultClass;
    }

    public <T> Mapper(Class<T> resultClass, Boolean strict) {
        this.resultClass = resultClass;
        this.strict = strict;
    }

    public <T> Mapper(Class<T> resultClass, Boolean strict, Boolean automap) {
        this.resultClass = resultClass;
        this.strict = strict;
        this.automap = automap;
    }

    public Mapper property(String property, String column) {
        mappings.put(column, new Mapping(property, column));
        return this;
    }

    public Mapper property(String property) {
        mappings.put(property, new Mapping(property, property));
        return this;
    }

    public Mapper property(String property, String column, int sqlType) {
        mappings.put(column, new Mapping(property, column).sqlType(sqlType));
        return this;
    }

    public Mapper property(String property, int sqlType) {
        mappings.put(property, new Mapping(property, property).sqlType(sqlType));
        return this;
    }

    public Mapper<T> join(String property, String column) {
        return add(property, column, MapType.JOIN_COLUMN);
    }

    public Mapper<T> join(String property) {
        return add(property, property, MapType.JOIN_COLUMN);
    }

    public Mapper<T> join(String property, String column, Integer sqlType) {
        return add(property, column, MapType.JOIN_COLUMN, sqlType);
    }

    public Mapper<T> join(String property, JoinTable joinTable, String column) {
        return add(new Mapping(property, column, MapType.JOIN_COLUMN).joinTable(joinTable), joinTable.getAlias() + "_" + column);
    }

    public Mapper<T> join(String property, JoinTable joinTable) {
        return join(property, joinTable, property);
    }

    public Mapper<T> join(String property, JoinTable joinTable, Mapper joinMap) {
        for (Mapping mapping : (Collection<Mapping>) joinMap.getMappings().values()) {
            if (mapping.getMapType() == MapType.PRIMARY_KEY)
                property(property + "." + mapping.getProperty(), joinTable.getAlias(), mapping.getSqlType());
            else if (mapping.getMapType() == MapType.TABLE_COLUMN)
                join(property + "." + mapping.getProperty(), joinTable, mapping.getColumn(), mapping.getSqlType());
        }
        return this;
    }

    public Mapper<T> join(String property, Mapper joinMap) {
        return join(property, joinMap, property);
    }

    public Mapper<T> join(String property, Mapper joinMap, String column) {
        JoinTable joinTable = new JoinTable(joinMap.getTablename(), column).on(tablename + "." + column + " = " + column + "." + joinMap.getIdMapping().getColumn());

        for (Mapping child : (Collection<Mapping>) joinMap.getMappings().values()) {
            if (child.getMapType() == MapType.PRIMARY_KEY) {
                Mapping mapping = new Mapping(property + "." + child.getProperty(), joinTable.getAlias(), MapType.TABLE_COLUMN);
                mapping.setSqlType(child.getSqlType());
                mapping.setJoinMap(joinMap);
                add(mapping, joinTable.getAlias());
            } else if (child.getMapType() == MapType.TABLE_COLUMN) {
                Mapping mapping = new Mapping(property + "." + child.getProperty(), child.getColumn(), MapType.JOIN_COLUMN);
                mapping.setSqlType(child.getSqlType());
                mapping.setJoinMap(joinMap);
                mapping.setJoinTable(joinTable);
                add(mapping, joinTable.getAlias() + "_" + child.getColumn());
            } else if (child.getMapType() == MapType.JOIN_COLUMN && child.getJoinMap() != null) {
                JoinTable childJoinTable = new JoinTable(child.getJoinTable().getName(), column + "_" + child.getJoinTable().getAlias())
                        .on(joinTable.getAlias() + "." + child.getJoinTable().getAlias() + " = " + column + "_" + child.getJoinTable().getAlias() + "." + child.getJoinMap().getIdMapping().getColumn());
                        
                join(property + "." + child.getProperty(), childJoinTable, child.getColumn(), child.getSqlType());
            }
        }
        return this;
    }

    public Mapping getIdMapping() {
        for (Mapping mapping : mappings.values())
            if (mapping.getMapType() == MapType.PRIMARY_KEY)
                return mapping;

        return null;
    }

    public List<Mapping> getIdMappings() {
        List<Mapping> idMappings = new ArrayList<>();
        for (Mapping mapping : mappings.values())
            if (mapping.getMapType() == MapType.PRIMARY_KEY)
                idMappings.add(mapping);

        return idMappings;
    }

    public Mapper<T> join(String property, JoinTable joinTable, String column, Integer sqlType) {
        return add(new Mapping(property, column, MapType.JOIN_COLUMN).sqlType(sqlType).joinTable(joinTable), joinTable.getAlias() + "_" + column);
    }

    public Mapper<T> add(String property, String column, MapType mapType) {
        mappings.put(column, new Mapping(property, column, mapType));
        return this;
    }

    public Mapper<T> add(String property, String column, MapType mapType, Integer sqlType) {
        mappings.put(column, new Mapping(property, column, mapType).sqlType(sqlType));
        return this;
    }

    public Mapper<T> id(String property, String column) {
        mappings.put(column, new Mapping(property, column, MapType.PRIMARY_KEY));
        return this;
    }

    public Mapper<T> id(String property, String column, Integer sqlType) {
        mappings.put(column, new Mapping(property, column, MapType.PRIMARY_KEY).sqlType(sqlType));
        return this;
    }

    public Mapper<T> id(String property, String column, String sequence) {
        return id(property, column, sequence, null);
    }

    public Mapper<T> id(String property, String column, String sequence, Integer sqlType) {
        mappings.put(column, new Mapping(property, column, MapType.PRIMARY_KEY).sqlType(sqlType).sequence(sequence));
        return this;
    }

    public Mapper<T> add(Mapping mapping, String column) {
        mappings.put(column, mapping);
        return this;
    }

    public Class getResultClass() {
        return resultClass;
    }

    public void setResultClass(Class<? extends T> resultClass) {
        this.resultClass = resultClass;
    }

    public Boolean getStrict() {
        return strict;
    }

    public void setStrict(Boolean strict) {
        this.strict = strict;
    }

    public Mapper strict() {
        this.strict = true;
        return this;
    }

    public Mapper lenient() {
        this.strict = false;
        return this;
    }

    public String getTablename() {
        return tablename;
    }

    public Mapper<T> tablename(String tablename) {
        this.tablename = tablename;
        return this;
    }

    public Boolean getAutomap() {
        return automap;
    }

    public void setAutomap(Boolean automap) {
        this.automap = automap;
    }

    public Mapper<T> automap(boolean automap) {
        setAutomap(automap);
        return this;
    }
}
