package no.tornado.query;

import no.tornado.query.typehandler.TypeHandler;
import no.tornado.query.typehandler.TypeMap;
import org.apache.commons.beanutils.BeanUtilsBean2;
import org.apache.commons.beanutils.NestedNullException;
import org.apache.commons.beanutils.PropertyUtilsBean;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.lang.reflect.InvocationTargetException;
import java.sql.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class Query<T> {
    private static Logger logger = LoggerFactory.getLogger(Query.class);

    public static final String SPACE = " ";
    public static final String SELECT = "SELECT";
    public static final String FROM = "FROM";
    public static final String WHERE = "WHERE";
    public static final String AND = "AND";
    public static final String OR = "OR";
    public static final String HAVING = "HAVING";
    public static final String GROUP = "GROUP";

    public static final List<String> DISCRIMINATORS = Arrays.asList(AND, OR);
    public static final List<String> NON_DISCRIMINATORS = Arrays.asList(HAVING, GROUP);

    private List<String> dynamics = new ArrayList<>();
    private List<Repeater> repeaters = new ArrayList<>();

    public static final ThreadLocal<Connection> connection = new ThreadLocal<>();
    private static final Pattern STATEMENT_PARAMETERS = Pattern.compile(":([a-zA-z0-9\\._-]+)");
    private static final Pattern COLLECTION_PARAMETER = Pattern.compile(".*:([a-zA-z0-9\\._-]+\\[\\]).*");
    public static final PropertyUtilsBean propertyUtils = BeanUtilsBean2.getInstance().getPropertyUtils();

    private StringBuilder statement = new StringBuilder();
    private Object parameters = new HashMap<String, Object>();
    private Map<String, Integer> parameterMappings = new HashMap<>();
    private Map<String, String> keyStatements = new HashMap<>();

    private Mapper<T> mapper;
    private ResultSetMetaData resultSetMetaData;
    private Map<String, Integer> columnIndexes;

    public Query(String sql) {
        add(sql);
    }

    public Query() {
    }

    public String toString() {
        return super.toString();
    }

    public static <T> Query<T> create(Class<? extends T> resultClass, String sql) {
        return create(resultClass).add(sql);
    }

    public static <T> Query<T> create(Class<? extends T> resultClass) {
        return new Query<T>().resultClass(resultClass);
    }

    public static <T> Query<T> create() {
        return new Query<T>();
    }

    public static <T> Query<T> create(String... sqls) {
        Query q = new Query<T>();
        for (String sql : sqls)
            q.add(sql);
        return q;
    }

    public Query<T> add(String... sqls) {
        for (String sql : sqls)
            add(sql);
        return this;
    }
    
    public Query<T> add(String sql) {
        if (sql == null || sql.length() == 0)
            return this;

        String nextElement = sql.substring(0, Math.max(0, sql.indexOf(SPACE))).toUpperCase();

        while (!dynamics.isEmpty() && !NON_DISCRIMINATORS.contains(nextElement))
            statement.append(SPACE).append(dynamics.remove(0));

        int length = statement.length();

        if (length > 0) {
            sql = sql.trim();
            if (sql.contains(SPACE)) {
                String lastElement = statement.toString().substring(Math.max(0, statement.lastIndexOf(SPACE) + 1)).toUpperCase();

                boolean removeNextElement = lastElement.equals(nextElement);
                if (!removeNextElement)
                    removeNextElement = lastElement.equals(WHERE) && DISCRIMINATORS.contains(nextElement);

                if (removeNextElement)
                    sql = sql.substring(nextElement.length()).trim();
            }

            statement.append(SPACE);
        }

        statement.append(sql);

        return this;
    }

    public Query<T> sequence(String property, String sequence) {
        return key(property, "SELECT nextval('" + sequence + "')");
    }

    public Query<T> key(String property, String sql) {
        keyStatements.put(property, sql);
        return this;
    }

    public Query<T> mapper(Mapper mapper) {
        this.mapper = mapper;
        return this;
    }

    public Query<T> resultClass(Class resultClass) {
        this.mapper = new Mapper(resultClass, false).lenient().automap(true);
        return this;
    }

    public Query<T> addIf(Boolean condition, String sql) {
        if (condition)
            add(sql);
        return this;
    }

    public Query<T> andIf(Boolean condition, String sql) {
        if (condition)
            add(AND + SPACE + sql);
        return this;
    }

    public Query<T> and(String sql) {
        return add(AND + SPACE + sql);
    }

    public Query<T> select(String sql) {
        statement.append(SELECT + SPACE + sql + SPACE);
        return this;
    }

    public static <E> Query<E> select(Mapper<E> mapper) {
        Query<E> q = (Query<E>) Query.create().mapper(mapper);
        String select = mapper.select();
        q.statement.append(select);
        return q;
    }

    public static <E> Query<E> byId(Mapper<E> mapper, Object id) {
        Query<E> q = (Query<E>) Query.create().mapper(mapper);
        String select = mapper.select();
        q.statement.append(select);
        q.where(mapper.getTablename() + "." + mapper.getIdMapping().getColumn() + " = " + ":id");
        q.param("id", id);
        return q;
    }

    public static <E> Query<E> by(Mapper<E> mapper, String by, Object... args) {
        Query<E> q = (Query<E>) Query.create().mapper(mapper);
        String select = mapper.select();
        q.statement.append(select);
        q.where();
        String[] bys = by.split("And");
        if (bys.length != args.length)
            throw new QueryException("Found " + bys.length + " property arguments but " + args.length + " actual arguments!");
        
        for (int i = 0; i < bys.length; i++) {
            String property = bys[i].substring(0, 1).toLowerCase() + bys[i].substring(1);
            Mapping mapping = mapper.getMappings().get(property);
            if (mapping == null)
                throw new QueryException(mapper.getClass().getSimpleName() + " does not contain a property called " + property);
            q.and(mapper.getTablename() + "." + mapping.getQualifiedColumn() + " = " + ":" + mapping.getQualifiedColumn());
            q.param(mapping.getQualifiedColumn(), args[i]);
        }
        return q;
    }

    public Query<T> select() {
        statement.append(SELECT + " * ");
        return this;
    }

    public Query<T> from(String sql) {
        statement.append(FROM + SPACE + sql);
        return this;
    }

    public Query<T> addEnclosed(String sql) {
        statement.append("(" + sql + ")");
        return this;
    }

    public Query<T> andEnclosed(String sql) {
        return add("AND (" + sql + ")");
    }

    public Query<T> andEnclosedIf(Boolean condition, String sql) {
        return addIf(condition, "AND (" + sql + ")");
    }

    public Query<T> leftJoin(String joinTable, String onStatement) {
        statement.append("LEFT JOIN " + joinTable + " ON " + onStatement);
        return this;
    }

    public Query<T> addUnless(Boolean condition, String sql) {
        if (!condition)
            add(sql);
        return this;
    }

    public Query<T> andUnless(Boolean condition, String sql) {
        if (!condition)
            add(AND + SPACE + sql);
        return this;
    }

    public Query<T> param(String name, Object param) {
        if (parameters instanceof Map)
            ((Map) parameters).put(name, param);
        else
            try {
                propertyUtils.setNestedProperty(parameters, name, param);
            } catch (Exception e) {
                throw new QueryException("Cannot set parameter " + name + " on parameter object " + parameters, e);
            }
        return this;
    }

    /**
     * Support for single object parameter
     */
    public Query<T> param(Object param) {
        this.parameters = param;
        return this;
    }

    public Query<T> param(String name, Object param, Integer targetSqlType) {
        if (name.contains("."))
            throw new RuntimeException("Cannot add nested param with targetSqlType, add mapping separately");

        param(name, param);
        paramMapping(name, targetSqlType);
        return this;
    }

    public Query<T> paramMapping(String name, Integer targetSqlType) {
        parameterMappings.put(name, targetSqlType);
        return this;
    }

    public PreparedStatement prepare() throws SQLException {
        String sql = statement.toString();
        if (!repeaters.isEmpty())
            for (int i = 0; i < repeaters.size(); i++)
                sql = sql.replace("#repeater" + (i + 1) + "#", processRepeater(repeaters.get(i)));

        sql = sql.replaceAll("\\s+", " ");

        if (logger.isDebugEnabled())
            logger.debug("Raw: " + sql);

        // Run key statements and apply properties - for now this does not accept parameters
        for (Map.Entry<String, String> keyEntry : keyStatements.entrySet()) {
            PreparedStatement stmt = connection.get().prepareStatement(keyEntry.getValue());
            ResultSet rs = stmt.executeQuery();
            if (rs.next()) {
                try {
                    Object resultObject = rs.getObject(1);
                    populateJavaObject(parameters, keyEntry.getKey(), resultObject);
                } catch (Exception e) {
                    throw new QueryException("Failed to set result from key query " + keyEntry.getKey(), e);
                }
            }
        }

        // Append parameters
        List<String> stmtParams = getStatementParameters(sql);
        sql = replacePlaceholders(sql);
        PreparedStatement stmt = connection.get().prepareStatement(sql);
        for (int i = 0; i < stmtParams.size(); i++) {
            String parameterName = stmtParams.get(i);
            int parameterIndex = i + 1;
            Integer targetSqlType = parameterMappings.get(parameterName);
            Object object;
            try {
                object = propertyUtils.getNestedProperty(parameters, parameterName);
            } catch (NestedNullException e) {
                object = null;
            } catch (Exception e) {
                throw new QueryException("Failed to retrieve property " + parameterName, e);
            }

            Object converted = object;

            try {
                Class javaClass = getPropertyType(parameters, parameterName);
                TypeHandler typeHandler = TypeMap.getHandler(javaClass, null);
                if (typeHandler != null) {
                    if (converted != null)
                        converted = typeHandler.convertParameter(converted);

                    if (targetSqlType == null)
                        targetSqlType = typeHandler.sqlType();
                }
            } catch (Exception ex) {
                ex.printStackTrace();
            }

            if (converted == null) {
                if (targetSqlType == null)
                    throw new QueryException("Don't know how to set " + parameterName + " to null, because no type can be inferred. Add typehandler or a parameter mapping for the property.");

                stmt.setNull(parameterIndex, targetSqlType);

            } else {
                if (targetSqlType != null)
                    stmt.setObject(parameterIndex, converted, targetSqlType);
                else
                    stmt.setObject(parameterIndex, converted);
            }
            
        }

        if (logger.isDebugEnabled())
            logger.debug(stmt.toString());

        return stmt;
    }

    private static String replacePlaceholders(String sql) {
        Matcher m = STATEMENT_PARAMETERS.matcher(sql);
        return m.replaceAll("?");
    }

    private List<String> getStatementParameters(String sql) {
        List<String> statementParameters = new ArrayList<>();
        Matcher m = STATEMENT_PARAMETERS.matcher(sql);
        while (m.find())
            statementParameters.add(m.group(1));
        return statementParameters;
    }

    private void populateJavaObject(Object javaObject, String property, Object resultObject) throws InvocationTargetException, NoSuchMethodException, IllegalAccessException {
        Class javaClass = getPropertyType(javaObject, property);
        if (javaClass == null) {
            try {
                propertyUtils.setNestedProperty(javaObject, property, resultObject);
            } catch (Exception ignored) {

            }
        } else {
            if (resultObject == null) {
                propertyUtils.setNestedProperty(javaObject, property, null);
            } else {
                TypeHandler handler = TypeMap.getHandler(javaClass, resultObject.getClass());
                if (handler == null) {
                    try {
                        propertyUtils.setNestedProperty(javaObject, property, resultObject);
                    } catch (Exception ex) {
                        if (mapper.getStrict())
                            throw new QueryException(ex);
                    }
                } else {
                    Object converted = handler.convertResult(resultObject);
                    propertyUtils.setNestedProperty(javaObject, property, converted);
                }
            }
        }
    }

    /**
     * Make sure the supplied object graph is populated by instantiating new objects along the path.
     */
    public static void ensureObjectsCreated(Object object, String property) throws InvocationTargetException, NoSuchMethodException, IllegalAccessException {
        if (property.contains(".")) {
            String firstLevel = property.substring(0, property.indexOf("."));
            Object created = createProperty(object, firstLevel);
            ensureObjectsCreated(created, property.substring(firstLevel.length() + 1));
        } else {
            createProperty(object, property);
        }
    }

    private static Object createProperty(Object object, String createProperty) throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
        Object propertyValue = propertyUtils.getNestedProperty(object, createProperty);
        if (propertyValue != null)
            return propertyValue;

        Class propertyClass = propertyUtils.getPropertyType(object, createProperty);
        if (propertyClass == null)
            throw new QueryException("Unable to instantiate parent property " + createProperty + " on " + object);

        try {
            Object created = propertyClass.newInstance();
            propertyUtils.setNestedProperty(object, createProperty, created);
            return created;
        } catch (InstantiationException e) {
            throw new QueryException("Unable to instantiate parent property " + createProperty + " on " + object, e);
        }
    }

    @SuppressWarnings("unchecked")
    private T convertRow(ResultSet rs) throws SQLException {
        if (mapper == null)
            mapper = new Mapper(HashMap.class).lenient().automap(true);

        try {
            T root = (T) mapper.getResultClass().newInstance();

            for (Mapping mapping : mapper.getMappings().values()) {
                try {
                    if (mapping.getSqlType() == null) {
                        Integer idx = getColumnIndexes(rs).get(mapping.getQualifiedColumn());
                        mapping.setSqlType(getResultSetMetaData(rs).getColumnType(idx));
                    }

                    Object resultObject = mapping.getResultObject(rs);
                    populateJavaObject(root, mapping.getProperty(), resultObject);
                } catch (Exception ex) {
                    if (mapper.getStrict())
                        throw ex;
                }
            }

            // Try to push values not specifically mentioned in mapper
            if (mapper.getAutomap()) {
                getResultSetMetaData(rs);

                int cc = resultSetMetaData.getColumnCount();
                for (int i = 0; i < cc; i++) {
                    int columnIndex = i + 1;
                    String columnName = resultSetMetaData.getColumnName(columnIndex);
                    if (!mapper.containsColumn(columnName)) {
                        Object resultObject = rs.getObject(columnIndex);
                        populateJavaObject(root, columnName, resultObject);
                    }
                }
            }

            return root;
        } catch (Exception ex) {
            throw new QueryException(ex);
        }
    }

    private Map<String, Integer> getColumnIndexes(ResultSet rs) throws SQLException {
        if (columnIndexes == null) {
            columnIndexes = new HashMap<>();
            for (int i = 0; i < getResultSetMetaData(rs).getColumnCount(); i++)
                columnIndexes.put(getResultSetMetaData(rs).getColumnName(i + 1).toLowerCase(), i + 1);
        }
        return columnIndexes;
    }

    private ResultSetMetaData getResultSetMetaData(ResultSet rs) throws SQLException {
        if (resultSetMetaData == null)
            resultSetMetaData = rs.getMetaData();
        
        return resultSetMetaData;
    }

    public List<T> rows() throws SQLException {
        List<T> rows = new ArrayList<>();
        ResultSet rs = prepare().executeQuery();
        while (rs.next())
            rows.add(convertRow(rs));

        rs.close();
        return rows;
    }

    public T first() throws SQLException {
        ResultSet rs = prepare().executeQuery();
        if (rs.next()) {
            T r = convertRow(rs);
            rs.close();
            return r;
        }
        return null;
    }

    public Integer delete() throws SQLException {
        PreparedStatement stmt = prepare();
        boolean isResultSet = stmt.execute();
        if (!isResultSet)
            return stmt.getUpdateCount();

        return 0;
    }

    public Integer insert() throws SQLException {
        PreparedStatement stmt = prepare();
        boolean isResultSet = stmt.execute();
        if (!isResultSet)
            return stmt.getUpdateCount();

        return 0;
    }

    public static Integer update(Mapper updateMap, Object param) throws SQLException {
        return new Query(updateMap.update(null)).param(param).update();
    }

    public static Integer delete(Mapper deleteMap, Object param) throws SQLException {
        return new Query(deleteMap.delete()).param(param).delete();
    }

    public static <T> Integer insert(Mapper<T> insertMap, Object param) throws SQLException {
        Query q = new Query<T>(insertMap.insert()).param(param);
        for (Mapping mapping : insertMap.getMappings().values())
            if (mapping.getSequence() != null)
                q.sequence(mapping.getProperty(), mapping.getSequence());
        
        return q.insert();
        
    }

    public Integer update() throws SQLException {
        PreparedStatement stmt = prepare();
        boolean isResultSet = stmt.execute();
        if (!isResultSet)
            return stmt.getUpdateCount();

        return 0;
    }

    public PreparedStatement execute() throws SQLException {
        PreparedStatement stmt = prepare();
        stmt.execute();
        return stmt;
    }

    public Query<T> where(String sql) {
        statement.append(SPACE + WHERE + SPACE + sql);
        return this;
    }

    public Query<T> where() {
        return dynamic(WHERE);
    }

    public Query<T> dynamic(String sql) {
        dynamics.add(sql);
        return this;
    }

    private Class getPropertyType(Object object, String property) throws InvocationTargetException, NoSuchMethodException, IllegalAccessException {
        if (property.contains("."))
            ensureObjectsCreated(object, property.substring(0, property.lastIndexOf(".")));

        Class propertyType = propertyUtils.getPropertyType(object, property);
        if (propertyType == null) {
            // Try to lookup type by actually getting the value, useful for determining types of map values and
            // other values where type cannot be determined by the owning class
            try {
                Object propertyValue = propertyUtils.getNestedProperty(object, property);
                if (propertyValue != null)
                    return propertyValue.getClass();
            } catch (Exception ignored) {

            }
        }
        return propertyType;
    }

    public Query<T> repeat(String conjunction, String sql) {
        return repeat(conjunction, "", sql);
    }

    public Query<T> repeat(String conjuntion, String prepend, String sql) {
        repeaters.add(new Repeater(conjuntion, prepend, sql));
        return add("#repeater" + repeaters.size() + "#");
    }

    private String processRepeater(Repeater repeater) {
        if (!(parameters instanceof Map))
            throw new QueryException("Parameters must be a map if you want to use repeat!");

        Matcher m = COLLECTION_PARAMETER.matcher(repeater.getSql());
        if (!m.matches())
            throw new QueryException("Tried to repeat without repeating on a parameter");

        String parameter = m.group(1).replaceAll("\\[\\]$", "");
        Collection collection;

        try {
            collection = (Collection) propertyUtils.getNestedProperty(parameters, parameter);
        } catch (IllegalAccessException | InvocationTargetException | NoSuchMethodException e) {
            return "";
        } catch (ClassCastException ex) {
            throw new QueryException("Tried to repeat on parameter " + parameter + ", which is not a collection", ex);
        }

        if (collection == null)
            return "";

        StringBuilder s = new StringBuilder(SPACE + repeater.getPrepend() + SPACE);

        Map params = (Map) parameters;
        Object[] elements = collection.toArray();
        for (int i = 0; i < collection.size(); i++) {
            if (i > 0)
                s.append(SPACE + repeater.getConjunction() + SPACE);

            String parameterName = parameter + i;
            params.put(parameterName, elements[i]);
            String entrySql = repeater.getSql().replaceAll(parameter + "\\[\\]", parameterName);
            s.append(entrySql);
        }

        return s.toString();
    }

}